"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.transformExportDeclaration = exports.getExported = exports.transformExportAssignment = void 0;
const ts = require("typescript");
const lua = require("../../../LuaAST");
const utils_1 = require("../../../utils");
const export_1 = require("../../utils/export");
const lua_ast_1 = require("../../utils/lua-ast");
const scope_1 = require("../../utils/scope");
const block_1 = require("../block");
const identifier_1 = require("../identifier");
const literal_1 = require("../literal");
const import_1 = require("./import");
exports.transformExportAssignment = (node, context) => {
    if (!context.resolver.isValueAliasDeclaration(node)) {
        return undefined;
    }
    const exportedValue = context.transformExpression(node.expression);
    // export = [expression];
    // ____exports = [expression];
    if (node.isExportEquals) {
        return lua.createVariableDeclarationStatement(lua_ast_1.createExportsIdentifier(), exportedValue, node);
    }
    else {
        // export default [expression];
        // ____exports.default = [expression];
        return lua.createAssignmentStatement(lua.createTableIndexExpression(lua_ast_1.createExportsIdentifier(), export_1.createDefaultExportStringLiteral(node)), exportedValue, node);
    }
};
function transformExportAllFrom(context, node) {
    utils_1.assert(node.moduleSpecifier);
    if (!context.resolver.moduleExportsSomeValue(node.moduleSpecifier)) {
        return undefined;
    }
    const moduleRequire = import_1.createModuleRequire(context, node.moduleSpecifier);
    const tempModuleIdentifier = lua.createIdentifier("____export");
    const declaration = lua.createVariableDeclarationStatement(tempModuleIdentifier, moduleRequire);
    const forKey = lua.createIdentifier("____exportKey");
    const forValue = lua.createIdentifier("____exportValue");
    const body = lua.createBlock([
        lua.createAssignmentStatement(lua.createTableIndexExpression(lua_ast_1.createExportsIdentifier(), forKey), forValue),
    ]);
    const pairsIdentifier = lua.createIdentifier("pairs");
    const forIn = lua.createForInStatement(body, [lua.cloneIdentifier(forKey), lua.cloneIdentifier(forValue)], [lua.createCallExpression(pairsIdentifier, [lua.cloneIdentifier(tempModuleIdentifier)])]);
    // Wrap this in a DoStatement to prevent polluting the scope.
    return lua.createDoStatement([declaration, forIn], node);
}
const isDefaultExportSpecifier = (node) => (node.name && node.name.originalKeywordKind === ts.SyntaxKind.DefaultKeyword) ||
    (node.propertyName && node.propertyName.originalKeywordKind === ts.SyntaxKind.DefaultKeyword);
function transformExportSpecifier(context, node) {
    const exportedSymbol = context.checker.getExportSpecifierLocalTargetSymbol(node);
    const exportedIdentifier = node.propertyName ? node.propertyName : node.name;
    const exportedExpression = literal_1.createShorthandIdentifier(context, exportedSymbol, exportedIdentifier);
    const isDefault = isDefaultExportSpecifier(node);
    const identifierToExport = isDefault
        ? export_1.createDefaultExportIdentifier(node)
        : identifier_1.transformIdentifier(context, node.name);
    const exportAssignmentLeftHandSide = export_1.createExportedIdentifier(context, identifierToExport);
    return lua.createAssignmentStatement(exportAssignmentLeftHandSide, exportedExpression, node);
}
function transformExportSpecifiersFrom(context, statement, moduleSpecifier, exportSpecifiers) {
    // First transpile as import clause
    const importClause = ts.createImportClause(undefined, ts.createNamedImports(exportSpecifiers.map(s => ts.createImportSpecifier(s.propertyName, s.name))));
    const importDeclaration = ts.createImportDeclaration(statement.decorators, statement.modifiers, importClause, moduleSpecifier);
    // Wrap in block to prevent imports from hoisting out of `do` statement
    const [block] = block_1.transformScopeBlock(context, ts.createBlock([importDeclaration]), scope_1.ScopeType.Block);
    const result = block.statements;
    // Now the module is imported, add the imports to the export table
    for (const specifier of exportSpecifiers) {
        result.push(lua.createAssignmentStatement(export_1.createExportedIdentifier(context, identifier_1.transformIdentifier(context, specifier.name)), identifier_1.transformIdentifier(context, specifier.name)));
    }
    // Wrap this in a DoStatement to prevent polluting the scope.
    return lua.createDoStatement(result, statement);
}
exports.getExported = (context, exportSpecifiers) => exportSpecifiers.elements.filter(exportSpecifier => context.resolver.isValueAliasDeclaration(exportSpecifier));
exports.transformExportDeclaration = (node, context) => {
    if (!node.exportClause) {
        // export * from "...";
        return transformExportAllFrom(context, node);
    }
    if (!context.resolver.isValueAliasDeclaration(node)) {
        return undefined;
    }
    if (ts.isNamespaceExport(node.exportClause)) {
        // export * as ns from "...";
        throw new Error("NamespaceExport is not supported");
    }
    const exportSpecifiers = exports.getExported(context, node.exportClause);
    // export { ... };
    if (!node.moduleSpecifier) {
        return exportSpecifiers.map(exportSpecifier => transformExportSpecifier(context, exportSpecifier));
    }
    // export { ... } from "...";
    return transformExportSpecifiersFrom(context, node, node.moduleSpecifier, exportSpecifiers);
};
//# sourceMappingURL=export.js.map